;######################################################################
;
; Copyright (C) 1999-2017, Michele Cappellari
; E-mail: cappellari_at_astro.ox.ac.uk
;
; For details on the method see:
;   Cappellari M., 2002, MNRAS, 333, 400
;
; Updated versions of the software are available from my web page
; http://purl.org/cappellari/software
;
; If you have found this software useful for your
; research, I would appreciate an acknowledgment to use of
; "the MGE fitting software developed by Cappellari (2002)"
;
; This software is provided as is without any warranty whatsoever.
; Permission to use, for non-commercial purposes is granted.
; Permission to modify for personal or internal use is granted,
; provided this copyright and disclaimer are included unchanged
; at the beginning of the file. All other rights are reserved.
;
;######################################################################
;+
; NAME:
;     SECTORS_PHOTOMETRY
;
; AUTHOR:
;       Michele Cappellari, Astrophysics Sub-department, University of Oxford, UK
;
; PURPOSE:
;       Perform photometry of a galaxy image along sectors equally spaced in
;       eccentric anomaly. This routine assumes four-fold symmetry, so measurements
;       in the four quadrants are averaged together. This routine is useful to
;       generate the input photometry required by the MGE fitting routine
;       MGE_FIT_SECTORS.
;
; EXPLANATION:
;       Further information on SECTORS_PHOTOMETRY is available in
;       Cappellari M., 2002, MNRAS, 333, 400
;
; CALLING SEQUENCE:
;       SECTORS_PHOTOMETRY, Img, Eps, Ang, Xc, Yc, Radius, Angle, Counts,
;           N_SECTORS=n_sectors, BADPIXELS=badpixels, MINLEVEL=minlevel, /PLOT
;
; INPUTS:
;       Img = The galaxy images as a 2D array.
;       Eps = The galaxy "average" ellipticity Eps = 1-b/a = 1-q'.
;           Photometry will be measured along elliptical annuli with
;           constant axial ellipticity Eps. The four quantities
;           (Eps,Ang,Xc,Yc) can be measured with the routine FIND_GALAXY.
;       Ang = Position angle measured counterclockwise from
;           the image Y axis, to the galaxy major axis.
;       Xc = X coordinate of the galaxy center in pixels.
;       Yc = Y coordinate of the galaxy center in pixels.
;
; OUTPUTS:
;       Radius = Vector containing the radius of the surface brightness
;               measurements, taken from the galaxy center. This is given
;               in units of PIXELS (!).
;       Angle = Vector containing the polar angle of the surface brightness
;               measurements, taken from the galaxy major axis.
;       Counts = Vector containing the actual surface brightness measurements
;               in COUNTS (!) at the polar coordinates specified by the vectors
;               Radius and Angle. These three vectors need to have the same
;               number of elements.
;
; OPTIONAL INPUT KEYWORDS:
;       N_SECTORS - Number of the sectors, equally spaced in eccentric anomaly,
;               from the galaxy major axis to the minor axis (one quadrant).
;               (default: 19 to cover the whole image with 5 degrees width).
;       BADPIXELS - Vector containing the indices of the pixels in the
;               array Img that have to be excluded from the photometry
;               calculation. Pixels indices can be selected using the
;               WHERE or POLYFILLV commands or interactively with
;               the DEFROI routine.
;       MINLEVEL - Scalar giving the minimum COUNTS level to include
;               in the photometry. The measurement along one profile
;               will stop when the counts go below this level.
;        PLOT - Set this keyword to show a checkerboard image, where
;               photometry was measured, at the end of the procedure.
;
; EXAMPLE:
;       The sequence of commands below was used to generate the complete
;       MGE model of the Figures 8-9 in Cappellari (2002).
;       1) The FITS file with the galaxy is read and the sky is subtracted;
;       2) the photometry along sectors is performed with SECTORS_PHOTOMETRY;
;       3) the resulting measurements are fitted with MGE_FIT_SECTORS;
;       4) the contour are printed on a PS file with MGE_PRINT_CONTOURS.
;       The geometric parameters of the galaxy (eps,ang,xc,yc) were
;       previously determined using FIND_GALAXY.
;
;           fits_read, 'ngc4342_f814w.fits', img, h
;
;           skylev = 0.55
;           img = img - skylev
;           scale = 0.0455  ; WFPC2/PC1 CCD pixel scale in arcsec/pixels
;
;           sigmaPSF = 0.8  ; Here I use one single Gaussian PSF
;           eps = 0.66      ; These values were measured with FIND_GALAXY
;           ang = 35.7
;           xc = 366
;           yc = 356
;           ngauss = 13
;
;           sectors_photometry, img, eps, ang, xc, yc, radius, angle, counts, $
;               MINLEVEL=0.2, /PLOT
;
;           MGE_fit_sectors, radius, angle, counts, eps, $
;               NGAUSS=ngauss, SIGMAPSF=sigmaPSF, SOL=sol, /PRINT, SCALE=scale
;
;           MGE_print_contours, img, ang, xc, yc, sol, BINNING=3, $
;               FILE='ngc4342.ps', SCALE=scale, MAGRANGE=9, SIGMAPSF=sigmaPSF
;
; PROCEDURES USED:
;       The following procedures are contained in the main SECTORS_PHOTOMETRY program.
;           MGE_DISPLAY                -- Show images even larger than the screen by resampling
;           MGE_COORDINATES            -- construct an image of coordinates
;           SECTOR_PHOTOMETRIC_PROFILE -- perform photometry along ONE position angle
;
;       Astronomy User's Library (http://idlastro.gsfc.nasa.gov/) routines needed:
;           BIWEIGHT_MEAN  -- Computes biweight mean of vectors.
;
; MODIFICATION HISTORY:
;       V1.0.0 First implementation for the NGC2681 photometric modeling.
;           Michele Cappellari, ESO Garching, 27 september 1999
;       V2.0.0 Major revisions, to use it with MGE_FIT_SECTORS.
;           Leiden, January 2000, MC
;       V2.1.0 Further updates, Padova, August 2000, MC
;       V2.1.1 Added compilation options, MC, Leiden 20 May 2002
;       V2.1.2 Allow for N_SECTORS=1 to get a single profile centered at
;           a given PA. Use biweight mean insted of sigma-clipped mean.
;           MC, Leiden, 30 April 2004
;       V2.1.3 Reduced amount of verbose output. MC, Leiden, 24 October 2004
;       V2.1.4 Replaced LOGRANGE keyword in example with the new MAGRANGE.
;           MC, Leiden, 1 May 2005
;       V2.1.5 Forces image to be positive when computing weighted radius
;           to prevent possible negative radii at large radii. Thanks to
;           Michael Williams for reporting the problem and the fix.
;           MC, Oxford, 16 February 2009
;       V2.1.6: Removed historic GOTO. MC, Oxford, 7 April 2017
;       V2.2.0: Sample image in eccentric anomaly rather than polar angle.
;         - Plot checkerboard image with /PLOT keyword.
;         - Stop profile if cnt <= 0.
;           MC, Oxford, 5 May 2017
;-
;----------------------------------------------------------------------------
PRO mge_display, img, ratio
COMPILE_OPT IDL2, HIDDEN

swx = !D.X_VSIZE    ; Window size in X in device units
swy = !D.Y_VSIZE    ; Size in Y
s = SIZE(img)
six = FLOAT(s[1])   ; Image size in X in pixels
siy = FLOAT(s[2])   ; Size in Y

ratio = siy/swy > six/swx
TVSCL, POLY_2D(img, [[0,0],[ratio,0]],[[0,ratio],[0,0]], 0, six/ratio, siy/ratio)

END
;--------------------------------------------------------------------
PRO mge_coordinates, q, pos_ang, xc, yc, n, rad, rell, ecc
COMPILE_OPT IDL2, HIDDEN

ang = (90 - pos_ang)/!RADEG

x = FINDGEN(n[0]) - xc
y = FINDGEN(n[1]) - yc
cap_meshgrid, x, y, xx, yy

x = xx*cos(ang) - yy*sin(ang)
y = xx*sin(ang) + yy*cos(ang)
x2 = x^2
y2 = y^2
rad = sqrt(x2 + y2)            ; Radius
rell = sqrt(x2 + y2/q^2)       ; Elliptical radius
ecc = atan(abs(y/q), abs(x))   ; Eccentric anomaly [0, pi/2]

END
;----------------------------------------------------------------------------
PRO sector_photometric_profile, data, grid, xc, yc, rad, rell, ecc, k, radius, cnt, $
    PLOT=plot, MINLEVEL=minlevel
COMPILE_OPT IDL2, HIDDEN

if ecc[xc,yc] NE -1 then ecc[xc,yc] = k  ; Always include central pixel unless bad
sector = WHERE(ecc EQ k)
irad = rell[sector]
h = HISTOGRAM(irad, REVERSE=ind) ; Get population and members of each "ellipse"
w = WHERE(h NE 0, m)

cnt = DBLARR(m)
radius = cnt

FOR j=0,N_ELEMENTS(w)-1 DO BEGIN
    sub = ind[ind[w[j]]:ind[w[j]+1]-1]    ; Find subscripts of members of region "r[j]"
    sub = sector[sub]                     ; Use the indices of the original image
    IF N_ELEMENTS(sub) GE 10 $
        THEN cnt[j] = BIWEIGHT_MEAN(data[sub]) $
        ELSE cnt[j] = MEAN(data[sub])
    IF cnt[j] LE minlevel THEN BEGIN
        cnt = cnt[0:j-1]
        radius = radius[0:j-1]
        BREAK
    ENDIF   ; Luminosity-weighted average radius in pixels
    radius[j] = TOTAL(rad[sub]*(data[sub]>0))/total(data[sub]>0)
    IF KEYWORD_SET(plot) THEN $
        IF (w[j] + k MOD 2) MOD 2 THEN grid[sub] = 0
ENDFOR

j = SORT(radius)
cnt = cnt[j]
radius = radius[j]

END
;----------------------------------------------------------------------------
PRO sectors_photometry, img, eps, ang, xc, yc, radius, angle, cnt, $
    BADPIXELS=badpixels, N_SECTORS=n_sectors, MINLEVEL=minlevel, PLOT=plot
;
; This routine performs photometry along sectors linearly spaced
; in angle between the major and minor axis of a galaxy.
; In output it returns the three vectors RADIUS, ANGLE, CNT,
; containing the photometric measurements in polar coordinates.
;
COMPILE_OPT IDL2
ON_ERROR, 2

IF N_ELEMENTS(n_sectors) EQ 0 THEN n_sectors = 19 ELSE n_sectors = n_sectors<37
IF N_ELEMENTS(minlevel) EQ 0 THEN minlevel = 0.0

s = SIZE(img)
xc = ROUND(xc)
yc = ROUND(yc)
q = 1 - eps
mge_coordinates, q, ang, xc, yc, s[1:2], rad, rell, ecc
rad[xc, yc] = 0.38  ; This is the average radius within the central pixel
rell[xc, yc] = 0.38

grid = img  ; This is only used for plotting

; Sample radii with 24 isophotes per decade: factor 1.1 spacing.
; Sample eccentric anomaly with n_sectors from 0-pi/2

rell = round(24.2*alog10(rell))
ecc = round(2*(n_sectors - 1)/!dpi*ecc)

if n_elements(badpixels) NE 0 then ecc[badpixels] = -1

PRINT, 'Measuring ', STRTRIM(n_sectors,2), ' profiles from 0 to 90 degrees...'

eccGrid = cap_range(0, !dpi/2, n_sectors)       ; Eccentric anomaly
angGrid = atan(tan(eccGrid)*q)*!RADEG           ; Polar angle

FOR k=0,n_sectors-1 DO BEGIN
    sector_photometric_profile, img, grid, xc, yc, rad, rell, ecc, k, $
        radius1, cnt1, MINLEVEL=minlevel, PLOT=plot
    IF k EQ 0 THEN BEGIN
        radius = radius1
        cnt = cnt1
        angle = REPLICATE(angGrid[k], N_ELEMENTS(radius1))
    ENDIF ELSE BEGIN
        radius = [radius, radius1]
        cnt = [cnt, cnt1]
        angle = [angle, REPLICATE(angGrid[k], N_ELEMENTS(radius1))]
    ENDELSE
ENDFOR

IF KEYWORD_SET(plot) THEN BEGIN
    LOADCT, 2, /SILENT
    mge_display, ALOG(grid > img[xc, yc]/1e4)
ENDIF

END
;----------------------------------------------------------------------------
